/**
 * ----------------------------------------------------------------------------
 *              BinHex for VAX/VMS, AXP/VMS, Ultrix (and whatever?)
 *              ===================================================
 * This program was created from 'mcvert': the authors and release notes for 
 * which are found below.  Without their work 'BinHex' would not exist.  
 * BinHex is offered publically on the same terms as mcvert.  Enjoy.

 * 9-AUG-93
 * DRASTICALLY changed 'mcvert' so that the program now functions as follows:
 *    Files are converted to/from the MAC file format used by Pacer (for VMS  
 * machines)  or CAP (for UNIX machines) from/to binhex or macbinary format.
 * The code for the encoding/decoding of the binhex and macbinary files 
 * is almost exactly the same as it was before.  The files in the 
 * Pacer/CAP format consist of data and/or resource forks and an info fork.
 * These files may be in the correct subdirectories, which are
 * [] ,[.afp_resource] and [.afp_resource.afp_infofile], respectively for 
 * VMS and ./, ./.resource and ./.finderinfo, respectively for UNIX.
 * If the directories are present, the forks are taken from or placed in the
 * correct subdirectories.  If not, the files are given appropriate 
 * extensions so that it is clear as to what their contents are.
 *    Not all of the variable and procedure names have been changed to reflect
 * the change in functionality of the program, so it may get (very) confusing  
 * at times.  
 *    Basically, what used to be the encoding (Download) of macbinary 
 * files is now the encoding into the Pacer/CAP format, and the the decoding
 * (Upload) converts the files to the binhex/macbinary format.
 *    Many compiler directives have been added so that the program functions 
 * similarly under VMS and UNIX (Ultrix to be exact). Some of the differences
 * are that routines such as mac2long, which change the byte-order of 
 * memory contents, are not defined under VMS, so they had to be simulated.
 * The header for the information fork has different formats under Pacer and 
 * CAP and these had to be taken into consideration.  Also, the routines to
 * determine whether or not the correct subdirectories are present are
 * OS dependent.
 *                  Suzy Gottesman: Senior Programmer (only programmer :-(
 *                   Research Computing Resource
 *                   NYU Medical Center, NY 10016.
 *                   <gottesman@mcclb0.med.nyu.edu>
 *                   <smith@mcclb0.med.nyu.edu> (backup)
 * ----------------------------------------------------------------------------
 * mcvert.c - version 1.05 - 10 January, 1990
 * Written by Doug Moore - Rice University - dougm@rice.edu - April '87

 * Sun bug fixes, assorted stuff - Jim Sasaki, March '89

 * Changed default max_line_size from 2000 to unlimited - Doug Moore, April, '89

 * Sun 3/60 doesn't like odd-sized structs.  Bug fixed - Doug Moore, April, '89
 *                                              - aided by Spencer W. Thomas

 * Didn't handle properly many hqx files combined in one file.  Bug fixed -
 *                                           Doug Moore, June, '89

 * Modified to handle MacBinaryII specification. Jim Van Verth, Sept, '89

 * Fixed a bug when there are blank lines in hqx data, as happens when newline
 * get translated to CRLF and then to \n\n, common for some file transfers.
 * The last hqx line would be lost if the previous line was blank or junk.
 *	Glenn Trewitt, Stanford University, 1990	(1.05)

 * Fixed a bug that occurred when extracting data or resource
 * forks on a Sun 4.  It was a byte alignment problem.
 * Rick Zaccone, zaccone@bucknell.edu.  April 1991.  Version 1.6

 * Fixed:
 *   Sent all "Converting ... " lines to stdout instead of stderr
 *   Changed mactypes.h for HP-UX systems
 *      Alan Danziger, aland@cs.brandeis.edu.  October 1991. Version 1.6.5

 * ----------------------------------------------------------------------------
 * External
 * -----
 * Fixed buffering bug when converting very small MacBinary files to hqx files.
 * Provide helpful usage line.
 * Control "Converting ... " lines separately with -S flag.
 * Make encoding and decoding consistent by ignoring locked and init flags.
 * Clean up some error messages; check for more errors; provide errno on error.
 * Updated the man page.
 * -----
 * Internal
 * -----
 * Reformat source (sorry, local standard used by tools is tab space == 3)
 * Remove compiler warning messages.
 * Rename some variables.
 * Added some comments to code.
 * Added some offsets to struct definitions.
 * Since the makefile has compilation flags,
 *    make the compiles depend on the Makefile.
 * -----
 * Thanks to all who have gone before for creating, maintaining,
 * improving, and providing this program and documentation.
 * -----
 * Joseph Skudlarek  Mentor Graphics  8005 SW Boeckman Rd  Wilsonville OR 97070
 * {Jskud@std.mentorg.com,Joseph_Skudlarek@mentorg.com}
 * ...{uunet,nosun,sequent,attunix,apollo}!mntgfx!Jskud
 * Version 1.70 09Jul92
 * ----------------------------------------------------------------------------

 * ----------------------------------------------------------------------------
 * External
 * -----
 * Added -V (Verbose) option (includes debugging information).
 * Fixed bug converting hqx to MacBinary if last line is ":".
 * Avoided a silent error and quick exit situation.
 * -----
 * Internal
 * -----
 * Got rid of almost all lint (SunOS and HP-UX) error messages.
 * Compiled on SunOs, HP-UX, DomainOS.
 * Incorporated Parag Patel <parag@netcom.com> changes for AU/X.
 *    Here's some diffs for really quick cheap hacks to get mcvert to compile
 *    and run under A/UX.  The main problem was that timeb does not exist, so
 *    I added 2 #ifdef TIMEVAL to use the System-V timeval package instead.
 *    The Makefile just has a -DTIMEVAL and a magic -U_SYSV_SOURCE to get
 *    around a pre-defined type "ulong" in sys/types.h (thanks to Apple).
 * Did more code overhauling:
 *    add lots more comments, rename variables, reformat source.
 * Put code in un_hqx to avoid suspected buffering problem.
 * -----
 * Joseph Skudlarek  Mentor Graphics  8005 SW Boeckman Rd  Wilsonville OR 97070
 * (503) 685-1576 (work)
 * {Jskud@std.MENTORG.Com,Joseph_Skudlarek@MENTORG.Com}
 * ...{uunet,nosun,sequent,attunix,apollo}!mntgfx!Jskud
 * Version 1.80 15Jul92
 * ----------------------------------------------------------------------------

 * ----------------------------------------------------------------------------
 * External
 * -----
 * Made hqx file scan processing much smarter
 *    so, for example, info-mac/comm/qwk-reader.hqx,
 *    complete with extraneous colons in column one, converts correctly
 *    (problem described by Edward John Sabol <es2j+@andrew.cmu.edu>)
 * Avoid silly perror on usage message (prompted by Edward John Sabol)
 * Improve error message regarding improper format
 * Added more caveats to man page
 * -----
 * Internal
 * -----
 * Fixed typo's in printf lines to pass all expected arguments
 *    (pointed out by Bo Holst-Christensen
 *    [holst@diku.dk/dikubhc1@uts.uni-c.dk/holst@login.dkuug.dk])
 * Tweak Makefile to ease shar creation and special case ulong, not A/UX
 * Add yet more comments and debugging code
 * -----
 * Joseph Skudlarek  Mentor Graphics  8005 SW Boeckman Rd  Wilsonville OR 97070
 * (503) 685-1576 (work)
 * {Jskud@std.MENTORG.Com,Joseph_Skudlarek@MENTORG.Com}
 * ...{uunet,nosun,sequent,attunix,apollo}!mntgfx!Jskud
 * Version 1.82 30Jul92
 * ----------------------------------------------------------------------------

 * ----------------------------------------------------------------------------
 * External
 * -----
 * relax exactly 64 characters per incoming hqx file, and
 * handle files without trailing newline, so, eg, 
 *    Telnet2.5docs.sit.hqx now converts correctly
 *       (failure reported by Justin Sullivan <justin@f2.facts.uky.edu>)
 *    now also processes info-mac/app/road-map.hqx correctly
 *       (failure reported by Victor Norton<norton@andy.bgsu.edu>)
 * rework the man page for improved clarity and completeness
 * -----
 * Internal
 * -----
 * avoid warning message from gcc on Sequent Balance mainframe
 *    reported by Justin Sullivan <justin@f2.facts.uky.edu>
 * bump max incoming line length to 2048 from 255
 * add mcvert.ps target to Makefile
 * -----
 * Joseph Skudlarek  Mentor Graphics  8005 SW Boeckman Rd  Wilsonville OR 97070
 * (503) 685-1576 (work)
 * {Jskud@std.MENTORG.Com,Joseph_Skudlarek@MENTORG.Com}
 * ...{uunet,nosun,sequent,attunix,apollo}!mntgfx!Jskud
 * Version 1.83 03Aug92
 * ----------------------------------------------------------------------------

 * ----------------------------------------------------------------------------
 * External
 * -----
 * Found and fixed problem with byte ordering plaguing users of
 *    Sequent's Balance running DYNIX, and DEC computers.
 *    The error message looked something like
 *       hqx_to_bin_fork: writing nnn too many bytes
 * Avoid generating debugging info if not to be printed -- cut runtime in half!
 * Generalize and incorporate -I (info) option processing provided by
 *    Paul Franklin, Computer Enginnering, Univ. of Calif., Davis CA 95616
 *    pdfranklin@ucdavis.edu.
 * Added heuristic to avoid false matches in mail headers.
 *    problem expertly characterized, solution beta tested, and subsequent
 *    improvement suggested by "Jim (J.) Lattanzi" <lattanzi@bnr.ca>
 *    so segmented comp.binaries.mac files (either multi-file or concatenated
 *    single file) should now convert correctly.
 * Added -H switch to disable heuristic.
 * Document heuristic in man page.
 * Fixed (long-standing) bug which precluded -p option from being recognized
 *    and verified decompressing and unpacking of PIT files working.
 *    Thanks to Dave Clemans for providing me with a version of PackIt.
 * Add the version to the extened Usage message emitted by the program.
 * Tune the syntax of the summary in the program and man page.
 * Cleaned up spelling mistakes in the man page.
 * -----
 * Internal
 * -----
 * Close all open streams --
 *    fix for binfile by Paul Franklin <pdfranklin@ucdavis.edu>
 * Incorporate changes suggested by Barry_Wolman@transarc.com
 *    to mactypes.h and Makefile for support of IBM RS/6000 running AIX 3.2
 *    reformat Makefile to avoid long option lines
 * Identify the right Makefile lines for Irix too
 *    suggested by Jack Repenning (jackr@dblues.wpd.sgi.com)
 * Clean up the stream handling and add mopen/mclose
 *    avoid unnecessary /dev/null opens
 *    all file open/close/read/write are checked for success
 * Lower lint content on SunOS and HP-UX.
 *    avoiding all "sometimes ignored" lint messages.
 * Improve modularity with mopen/mclose/converting routines.
 * Tune debugging output information.
 * Verify that passes smoke tests on DomainOS/SunOS/HP-UX/ULTRIX.
 * Reformat these comments to avoid tabs.
 * -----
 * Joseph Skudlarek  Mentor Graphics  8005 SW Boeckman Rd  Wilsonville OR 97070
 * (503) 685-1576 (work)
 * {Jskud@std.MENTORG.Com,Joseph_Skudlarek@MENTORG.Com}
 * ...{uunet,nosun,sequent,attunix,apollo}!mntgfx!Jskud
 * Version 1.87 25Sep92
 * ----------------------------------------------------------------------------

 * ----------------------------------------------------------------------------
 * External
 * -----
 * Add README file which describes how to configure and compile mcvert
 * Handle multiple BinHex4.0 inputs in a single file again (thanks to
 *    <Mark_Larimer@pigeon.cpg.cdc.com> for pointing out this regression)
 * Emit the MacBinary header if verbose (to get create and modify times)
 * -----
 * Internal
 * -----
 * Rename some variables, create mac2unix (time) routine, more comments
 * Keep lint content low
 * Pull unnecessary include of <net/nh.h> -- avoid breaking AIX 3.1
 * Fix = vs == typo dealing with protect bit
 * -----
 * Joseph Skudlarek  Mentor Graphics  8005 SW Boeckman Rd  Wilsonville OR 97070
 * (503) 685-1576 (work)
 * {Jskud@std.MENTORG.Com,Joseph_Skudlarek@MENTORG.Com}
 * ...{uunet,nosun,sequent,attunix,apollo}!mntgfx!Jskud
 * Version 1.88 08Dec92
 * ----------------------------------------------------------------------------

 * ----------------------------------------------------------------------------
 * External
 * -----
 * emit input file names when verbose is enabled
 *    (suggested by franklin@eecs.ucdavis.edu)
 * make it easier to build on AT&T 3B2's
 * -----
 * Internal
 * -----
 * provide compile time switch to avoid bzero and bcopy, and use memset
 *    and memcpy instead (pointed out by linger@drystone.attmail.com, and
 *    requested again by Larry S. Staples <attjp4!lss>)
 * update Makefile to include incantations required for AT&T 3B2's
 * fflush all diagnostic output to ensure correct order when output to a file
 * -----
 * Joseph Skudlarek  Mentor Graphics  8005 SW Boeckman Rd  Wilsonville OR 97070
 * (503) 685-1576 (work)
 * {Jskud@std.MENTORG.Com,Joseph_Skudlarek@MENTORG.Com}
 * ...{uunet,nosun,sequent,attunix,apollo}!mntgfx!Jskud
 * Version 1.89 05Jan93
 * ----------------------------------------------------------------------------

 * ----------------------------------------------------------------------------
 * External
 * -----
 * incorporate MAC_FILETYPE support provided by <root@genome.stanford.edu>
 * minor edits to man page
 * -----
 * Internal
 * -----
 * update Makefile to simplify incantations required for AT&T 3B2's
 * -----
 * Joseph Skudlarek  Mentor Graphics  8005 SW Boeckman Rd  Wilsonville OR 97070
 * (503) 685-1576 (work)
 * {Jskud@std.MENTORG.Com,Joseph_Skudlarek@MENTORG.Com}
 * ...{uunet,nosun,sequent,attunix,apollo}!mntgfx!Jskud
 * Version 1.90 04Mar93
 * ----------------------------------------------------------------------------

 * ----------------------------------------------------------------------------
 * External
 * -----
 * handle -b (both .data and .rsrc at same time) option
 *    ( -> MacBinary suggested by bb@math.ufl.edu)
 * remove anomalous file extension handling
 *    (suggested by bb@math.ufl.edu)
 * regularize MAC_EDITOR (author) and MAC_FILETYPE (file type) handling
 * detect and report important file format errors
 * emit output file name too
 * revise Usage line
 * massively revise man page to reflect changes and generally overhaul
 * -----
 * Internal
 * -----
 * avoid overwriting internal storage (what a chore!)
 *    for example, mcvert -UI *.hqx used to abort with a segmentation violation
 *    symptom reported by franklin@eecs.ucdavis.edu  Thu Sep 24 16:39:21 1992
 * check return values from all getc/putc operations
 * find and fix ancient bug extracting resource fork from MacBinary format
 * identify failing file in EOF error messages
 * clarify and amend distribution and update restrictions
 * expand disclaimer (patterned after INFO-MAC CD-ROM -- Thx, Cliff and Joe!)
 * -----
 * Joseph Skudlarek  Mentor Graphics  8005 SW Boeckman Rd  Wilsonville OR 97070
 * (503) 685-1576 (work)
 * {Jskud@std.MENTORG.Com,Joseph_Skudlarek@MENTORG.Com}
 * ...{uunet,nosun,sequent,attunix,apollo}!mntgfx!Jskud
 * Version 2.00 28Feb93
 * ----------------------------------------------------------------------------

 * ----------------------------------------------------------------------------
 * This program may be freely distributed for non-profit purposes.  It
 * may not be sold, by itself or as part of a collection of software.
 * However, it may be distributed in source form with large
 * collections of freeware and shareware, such as the INFO-MAC CD-ROM,
 * which charge only a modest fee for publishing, but not selling, the
 * software.  It may be freely modified as long as no modified version
 * is independently distributed.  Modifications of interest to all can
 * be incorporated into the program by sending them to me for
 * inclusion and redistribution, or by releasing an updated mcvert to
 * the info-mac archives.  Parts of the code can be used in other
 * programs.  We hope you find mcvert useful, and enjoy using it.
 * ----------------------------------------------------------------------------
 * ----------------------------------------------------------------------------
 * DISCLAIMER -- USE mcvert software AT YOUR OWN RISK
 * This mcvert software is provided "as is" without warrantee of any
 * kind.  The entire risk as to the results and performance of the
 * software is assumed by you, and in no event will we be liable for
 * any consequential, incidental, or indirect damages suffered in the
 * course of using this software.
 * ----------------------------------------------------------------------------

 * ----------------------------------------------------------------------------
 * Things that yet could be done:
 * ---
 * tune unixify routine (replace '~' only if first), unixify all output files
 * check for more file format errors, eg, MacBinary must be multiple of 128
 * create and call macify routine (replace ':', for example)
 * check return values from fputs/fprintf
 * provide header heuristic tuning option: set length and/or same sensitivity
 * ----------------------------------------------------------------------------
 */

/*
 * Naming
 *		DOWNLOAD
 *			=> converting TO Pacer/CAP
 *			=> direction == FORWARDS
 *			=> use un_* routines (un => UNdo encoding?)
 *		UPLOAD
 *			=> converting FROM Pacer/CAP
 *			=> direction == BACKWARDS
 *			=> use re_* routines (re => Really Encode?)
 */

#include "mactypes.h"

/* it would be natural to use an enum here, but avoid "fancy" features */

#define FORWARDS 0
#define BACKWARDS 1

FILE *verbose;
FILE *debug;
FILE *convert;
FILE *devnull;
FILE *output;

char **hqxnames, **hqxnames_left;
char *dir, *ext, *mac_auth, *mac_type,*infodir,*rsrcdir;
int translate_eol,have_subdirs,orig_filename;
char *maxlines_str;
int maxlines;

/* used to skip suspect mail header lines */
int suspect_shorter = 12;
int suspect_same = 1;

/* used to avoid writing output files */
int info_only;

char Usage[] = "\
Usage: %s { [option] ... name ...} ...\n\
 version:\t%4.2f\n\
 default:\t-xD\n\
\n\
 option:\n\
\t-x\tBinHex        .hqx  <-> Pacer/CAP\n\
\t-m\tMacbinary        .bin  <-> Pacer/CAP\n\
\n\
\t-D\tDownload        .hqx/.bin -> Pacer/CAP\n\
\t-U\tUpload        PAcer/CAP -> .hqx/.bin\n\
\n\
Environment:\n\
\tMAC_FILETYPE  \tTEXT|????\tMac file type for Text|other\n\
\tMAC_DLOAD_DIR \t.        \tdirectory for -D\n\
\tMAC_LINE_LIMIT\tnone     \tmaximum line length for -Ux\n\
";

char *cmdname;

main(argc, argv)
	int argc;
	char **argv;
{
	char *flags, *getenv();
	int direction, mode, unpit_flag,found_rsrc,found_info;
#ifdef VMS
      	int  status,context;
      	static $DESCRIPTOR(rsrc_fname,
                "[]afp_resource.dir");
      	static $DESCRIPTOR(rsrc_ret,
                "[]afp_resource.dir");
      	static $DESCRIPTOR(info_fname,
                "[.afp_resource]afp_infofile.dir");
      	static $DESCRIPTOR(info_ret,
                "[.afp_resource]afp_infofile.dir");
#else
      	DIR *dirp;
      	struct direct *dp;
#endif
	cmdname = argv[0];

	/* Early error and clean exit if missing arguments */
	if (argc < 2) {
		usage1();
		/*NOTREACHED*/
	}

	devnull = fopen("/dev/null", "w+");

	argv++;
	argc--;
	verbose = stderr;
	debug = devnull;
	output = stdout;
	convert = stdout;
	direction = FORWARDS;
	mode = HQX;
	unpit_flag = 0;

	mac_type = getenv("MAC_FILETYPE");
	mac_auth = getenv("MAC_EDITOR");

	if ((ext = getenv("MAC_EXT")) == NULL)
		ext = ".bin";
	if ((dir = getenv("MAC_DLOAD_DIR")) == NULL)
		dir = ".";

      	orig_filename = FALSE;
      	found_rsrc = FALSE;
      	found_info = FALSE;
#ifdef VMS
      	status = lib$find_file(&rsrc_fname,&rsrc_ret,&context,0,0,0,0);
      	if (status == RMS$_NORMAL) { 
      	   status = lib$find_file(&info_fname,&info_ret,&context,0,0,0,0);
      	   if (status == RMS$_NORMAL) {
      			found_rsrc = TRUE;
      			found_info = TRUE;
      			rsrcdir = "afp_resource";
      			infodir = "afp_infofile";
      	   }
      	}
#else
      		rsrcdir = ".resource";
      		infodir = ".finderinfo";
      	dirp = opendir(".");
      	for (dp = readdir(dirp); dp != NULL; dp = readdir(dirp)) {
      		if (!strcmp(dp->d_name,rsrcdir))
      			found_rsrc = TRUE;
      		if (!strcmp(dp->d_name,infodir))
      			found_info = TRUE;
      	}
      	closedir(dirp);
#endif
      	if (found_rsrc && found_info) 
      		have_subdirs = TRUE;
      	else
      		have_subdirs = FALSE;      			
	if ((maxlines_str = getenv("MAC_LINE_LIMIT")) == NULL)
		maxlines = 0;
	else {
		maxlines = atoi(maxlines_str);
		if (maxlines < MIN_HQX_LINES) {
			fprintf(stderr, "%s: %s; was %d; reset to %d\n",
				cmdname,
				"warning: MAC_LINE_LIMIT too small",
				maxlines, MIN_HQX_LINES);
			fflush(stderr);
			maxlines = MIN_HQX_LINES;
		}
	}

	/* Make command line arguments globally accessible */
	hqxnames = (char **) calloc((unsigned)argc + 1, sizeof(char *));
	hqxnames_left = hqxnames;
	while (argc--)
		*hqxnames_left++ = *argv++;

	/* Flag the end of the list */
	*hqxnames_left = "-";
	hqxnames_left = hqxnames;

	/* While not at the end of the list */
	while (strcmp(*hqxnames_left, "-")) {
		translate_eol = 0;
		if (hqxnames_left[0][0] == '-') {
			flags = *hqxnames_left++;
			while (*++flags)
				switch (*flags) {
				case 'f':
      				case 'F':
					orig_filename = TRUE;
					break;
				case 'm':
      				case 'M':
					mode = MACBIN;
					break;
				case 'x':
				case 'X':
					mode = HQX;
					break;
				case 'd':
				case 'D':
					direction = FORWARDS;
					break;
				case 'u':
				case 'U':
					direction = BACKWARDS;
					break;
/*
				case 'p':
					unpit_flag = 1;
					break;
				case'H':
					suspect_shorter = suspect_same = 0;
					break;
*/
				default:
					usage1();
					/*NOTREACHED*/
				}
		}

		if (direction == BACKWARDS)
      			re_hqx(mode);
		else 
			un_hqx(unpit_flag,mode);
	}

	exit(0);
	/*NOTREACHED*/
}

/* An array useful for CRC calculations that use 0x1021 as the "seed" */
word magic[] = {
    0x0000, 0x1021, 0x2042, 0x3063, 0x4084, 0x50a5, 0x60c6, 0x70e7,
    0x8108, 0x9129, 0xa14a, 0xb16b, 0xc18c, 0xd1ad, 0xe1ce, 0xf1ef,
    0x1231, 0x0210, 0x3273, 0x2252, 0x52b5, 0x4294, 0x72f7, 0x62d6,
    0x9339, 0x8318, 0xb37b, 0xa35a, 0xd3bd, 0xc39c, 0xf3ff, 0xe3de,
    0x2462, 0x3443, 0x0420, 0x1401, 0x64e6, 0x74c7, 0x44a4, 0x5485,
    0xa56a, 0xb54b, 0x8528, 0x9509, 0xe5ee, 0xf5cf, 0xc5ac, 0xd58d,
    0x3653, 0x2672, 0x1611, 0x0630, 0x76d7, 0x66f6, 0x5695, 0x46b4,
    0xb75b, 0xa77a, 0x9719, 0x8738, 0xf7df, 0xe7fe, 0xd79d, 0xc7bc,
    0x48c4, 0x58e5, 0x6886, 0x78a7, 0x0840, 0x1861, 0x2802, 0x3823,
    0xc9cc, 0xd9ed, 0xe98e, 0xf9af, 0x8948, 0x9969, 0xa90a, 0xb92b,
    0x5af5, 0x4ad4, 0x7ab7, 0x6a96, 0x1a71, 0x0a50, 0x3a33, 0x2a12,
    0xdbfd, 0xcbdc, 0xfbbf, 0xeb9e, 0x9b79, 0x8b58, 0xbb3b, 0xab1a,
    0x6ca6, 0x7c87, 0x4ce4, 0x5cc5, 0x2c22, 0x3c03, 0x0c60, 0x1c41,
    0xedae, 0xfd8f, 0xcdec, 0xddcd, 0xad2a, 0xbd0b, 0x8d68, 0x9d49,
    0x7e97, 0x6eb6, 0x5ed5, 0x4ef4, 0x3e13, 0x2e32, 0x1e51, 0x0e70,
    0xff9f, 0xefbe, 0xdfdd, 0xcffc, 0xbf1b, 0xaf3a, 0x9f59, 0x8f78,
    0x9188, 0x81a9, 0xb1ca, 0xa1eb, 0xd10c, 0xc12d, 0xf14e, 0xe16f,
    0x1080, 0x00a1, 0x30c2, 0x20e3, 0x5004, 0x4025, 0x7046, 0x6067,
    0x83b9, 0x9398, 0xa3fb, 0xb3da, 0xc33d, 0xd31c, 0xe37f, 0xf35e,
    0x02b1, 0x1290, 0x22f3, 0x32d2, 0x4235, 0x5214, 0x6277, 0x7256,
    0xb5ea, 0xa5cb, 0x95a8, 0x8589, 0xf56e, 0xe54f, 0xd52c, 0xc50d,
    0x34e2, 0x24c3, 0x14a0, 0x0481, 0x7466, 0x6447, 0x5424, 0x4405,
    0xa7db, 0xb7fa, 0x8799, 0x97b8, 0xe75f, 0xf77e, 0xc71d, 0xd73c,
    0x26d3, 0x36f2, 0x0691, 0x16b0, 0x6657, 0x7676, 0x4615, 0x5634,
    0xd94c, 0xc96d, 0xf90e, 0xe92f, 0x99c8, 0x89e9, 0xb98a, 0xa9ab,
    0x5844, 0x4865, 0x7806, 0x6827, 0x18c0, 0x08e1, 0x3882, 0x28a3,
    0xcb7d, 0xdb5c, 0xeb3f, 0xfb1e, 0x8bf9, 0x9bd8, 0xabbb, 0xbb9a,
    0x4a75, 0x5a54, 0x6a37, 0x7a16, 0x0af1, 0x1ad0, 0x2ab3, 0x3a92,
    0xfd2e, 0xed0f, 0xdd6c, 0xcd4d, 0xbdaa, 0xad8b, 0x9de8, 0x8dc9,
    0x7c26, 0x6c07, 0x5c64, 0x4c45, 0x3ca2, 0x2c83, 0x1ce0, 0x0cc1,
    0xef1f, 0xff3e, 0xcf5d, 0xdf7c, 0xaf9b, 0xbfba, 0x8fd9, 0x9ff8,
    0x6e17, 0x7e36, 0x4e55, 0x5e74, 0x2e93, 0x3eb2, 0x0ed1, 0x1ef0
};


/*
 * calc_crc() --
 *   Compute the MacBinary II-style CRC for the data pointed to by p, with the
 *   crc seeded to seed.
 *
 *   Modified by Jim Van Verth to use the magic array for efficiency.
 */
short
calc_mb_crc(p, len, seed)
	unsigned char *p;
	long len;
	short seed;
{
	short hold;		/* crc computed so far */
	long i;			/* index into data */

	extern unsigned short magic[];	/* the magic array */

	hold = seed;			       /* start with seed */
	for (i = 0; i < len; i++, p++) {
		hold ^= (*p << 8);
		hold = (hold << 8) ^ magic[(unsigned char) (hold >> 8)];
	}

	return (hold);
}				/* calc_crc() */


/* Report a fatal error, and exit, never return */
error(msg, name)
	char msg[], name[];

{
	fprintf(stderr, msg, name);
	(void)putc('\n', stderr);
	perror("\nlast perror (may not be relevant)");
	fprintf(stderr, "%s: exiting\n", cmdname);
	if (debug == stderr)
		abort();
	exit(2);
	/*NOTREACHED*/
}

/* replace illegal Unix characters in file name */
/* make sure host file name doesn't get truncated beyond recognition */
unixify(np)
	register char *np;
{
	register ulong c;

	c = strlen(np);
	if (c > SYSNAMELEN - MAXEXTENSION)
		c = SYSNAMELEN - MAXEXTENSION;
	np[c] = '\0';
	np--;
	while (c = *++np)
		if (c <= ' ' || c == '/' || c > '~')
			*np = '_';
}

/* replace illegal VMS characters in file name */
/* make sure host file name doesn't get truncated beyond recognition */

vaxify(np)
	register char *np;
{
	register ulong c,c_ext;
      	char 	*dot,*start;


      	if (dot = strrchr(np,'.')) {
      		c_ext = strlen(dot+1);
      		if (c_ext > 32) 
      			c_ext = 32;
      		*(dot+c_ext+1) = '\0';
      		c = strlen(np) - c_ext - 1;
      		if (c>32) {
      			c = 32;
            		strncpy(np+c,dot,strlen(dot));
      			*(np+c+strlen(dot)) = '\0';
      		}
      	}
      	else {
      		c = strlen(np);
      		if (c > 32)
			c = 32;
      		np[c] = '\0';
      	}
      	start = np;
      	np--;
	while (c = *++np) 
		if (c<='#' || (c>='%' && c < '-') || (c=='-' && np == start)||
      			(c=='.' && np != dot) ||  (c== '/') || 
      			(c >= ':' && c<='@') || 
                        (c >= '[' && c<'a') || c>='{') 
      				*np = '_';
}





/*
 * Unix time (GMT since 1-1-1970)
 * Mac time (local since 1-1-1904)
 */
#define MACTIMEDIFF 0x7c25b080	/* Mac time of 00:00:00 GMT, Jan 1, 1970 */

/* Convert Unix time to Mac time */
ulong
unix2mac(xtime)
	ulong xtime;
{
#ifdef TIMEVAL
	struct timeval t;
	struct timezone tz;

	gettimeofday(&t, &tz);
	return long2mac(xtime + MACTIMEDIFF
	   - 60 * (tz.tz_minuteswest - 60 * tz.tz_dsttime));

#else
      	byte    temp[4],temp0[4];
	struct timeb tp;
      	ulong temptime;

	ftime(&tp);
      	temptime = xtime + MACTIMEDIFF - 60 * (tp.timezone - 60 * tp.dstflag);

#ifdef VMS
      	temptime = xtime + MACTIMEDIFF - 60 * (tp.timezone - 60 * tp.dstflag);
      	bcopy((char*)&temptime, (char*)temp0, 4);
      	bcopy((char*)(temp0+3), (char*)temp, 1);
      	bcopy((char*)(temp0+2), (char*)(temp+1) ,1);
      	bcopy((char*)(temp0+1), (char*)(temp+2) ,1);
      	bcopy((char*)temp0, (char*)(temp+3), 1);
      	bcopy((char*)temp, (char*)&temptime, 4);
      	return(temptime);
#else
	return long2mac(xtime + MACTIMEDIFF
	    - 60 * (tp.timezone - 60 * tp.dstflag));
#endif
#endif
}

/* Convert Mac time to Unix time */
ulong
mac2unix(xtime)
	ulong xtime;
{
#ifdef TIMEVAL
	struct timeval t;
	struct timezone tz;

	gettimeofday(&t, &tz);
	return (mac2long(xtime) - MACTIMEDIFF
	   + 60 * (tz.tz_minuteswest - 60 * tz.tz_dsttime));
#else
	struct timeb tp;
      	byte    temp[4],temp0[4];
      	ulong temptime;

	ftime(&tp);

#ifdef VMS
      	bcopy((char*)&xtime, (char*)temp0, 4);
      	bcopy((char*)(temp0+3), (char*)temp, 1);
      	bcopy((char*)(temp0+2), (char*)(temp+1) ,1);
      	bcopy((char*)(temp0+1), (char*)(temp+2) ,1);
      	bcopy((char*)temp0, (char*)(temp+3), 1);
      	bcopy((char*)temp, (char*)&temptime, 4);
	return (temptime - MACTIMEDIFF
	    + 60 * (tp.timezone - 60 * tp.dstflag));
#else
	return (mac2long(xtime) - MACTIMEDIFF
	    + 60 * (tp.timezone - 60 * tp.dstflag));
#endif
#endif
}


usage1()
{
	fprintf(stderr, Usage, cmdname, VERSION/100.);
	exit(1);
	/*NOTREACHED*/
}


/* My FileIO routines, to enable clean implementation of info_only */
/* If info_only and open for write, inform user and use devnull instead */
/* If trying to close devnull, don't */

FILE *
mopen(name, exten, type)
	char *name;
	char *exten;
	char *type;
{
	FILE *result;


	switch (*type) {
	case 'r':
		result = fopen(name, type);
		if (result == NULL && *exten) {
			/* see if adding the extension would help */
			int len_root = strlen(name);
			int len_ext = strlen(exten);
			char *dotspot = name + len_root - len_ext;
			if (strcmp(exten, dotspot)) {
				if (len_root + len_ext >= SYSNAMELEN)
					error("Error: generated file name would be too long", "");
				strcat(name, exten);
				result = fopen(name, type);
			}
		}
/*
		if (result == NULL)
			error("Cannot open %s for read", name);
		else {
			fprintf(verbose, "\n%-15s%-27s\n",
				"Input file", name);
			fflush(verbose);
		}
*/
		if (result != NULL) {
			fprintf(verbose, "\n%-15s%-27s\n",
				"Input file", name);
			fflush(verbose);
		}
		break;

	case 'w':
		if (info_only) {
			fprintf(verbose, " %-14s%-27s -I specified\n",
				" No output to", name);
			fflush(verbose);
			result = devnull;
		} else {
#ifdef VMS
/*
			result = fopen(name, type,"mrs=128","rfm=fix");
*/
			result = fopen(name, type);
#else
			result = fopen(name, type);
#endif
			if (result == NULL)
				error("Cannot open %s for write", name);
			else {
				fprintf(verbose, "%-15s%-27s\n",
					"Output file", name);
				fflush(verbose);
			}
		}
		break;

	default:
		fprintf(stderr, "%s: internal error in mopen -- exiting\n", cmdname);
		exit(2);

	}

	return result;
}

mclose(stream_p, name)
	FILE **stream_p;
	char *name;
{
#ifdef VMS
	if (*stream_p && *stream_p != devnull) {
		if (fclose(*stream_p) == EOF)
			error("Error closing %s", name);
	}
#else
	if (*stream_p && *stream_p != devnull) {
		if (fclose(*stream_p) == EOF)
			error("Error closing %s", name);
	}
#endif
}

converting(name, type, auth)
	char *name;
	byte *type;
	byte *auth;
{

	fprintf(convert,
	    "%-15s%-30s type = \"%4.4s\", author = \"%4.4s\"\n",
 	    info_only ? "Processing" : "Converting",
		 name, type, auth);
	fflush(convert);
}

print_bin_hdr(msg, hdr)
	char *msg;
	info_header *hdr;
{
	ulong otime, xtime;

	bcopy((char*)hdr->ctim, (char*)&otime, 4);

printf("otime = %d\n",otime);
/*	DEBUG && fprintf(debug, "DEBUG: verifying mac2unix/unix2mac: 0x%8lx, 0x%8lx\n",
		otime, unix2mac(mac2unix(otime)));
*/
printf("DEBUG: verifying mac2unix/unix2mac: 0x%8lx, 0x%8lx\n",
		otime, unix2mac(mac2unix(otime)));
	DEBUG && fprintf(debug, "\
%s\n\
\tName     %.*s\n\
\tType     %.4s\n\
\tCreator  %.4s\n\
",
		msg,
		hdr->nlen, hdr->name,
		hdr->type,
		hdr->auth);

	bcopy((char*)hdr->ctim, (char*)&otime, 4);
	xtime = mac2unix(otime);
/*
	DEBUG && fprintf(debug, "raw ctim: 0x%8lx, mac2unix: 0x%8lx\n",
		otime, xtime);
*/
printf("raw ctim: 0x%8lx, mac2unix: 0x%8lx\n",
		otime, xtime);
	DEBUG && fflush(debug);

	fprintf(verbose, "\tCreated  %s", ctime((time_t *)&xtime));
	fflush(verbose);

	bcopy((char*)hdr->mtim, (char*)&otime, 4);
	xtime = mac2unix(otime);
	DEBUG && fprintf(debug, "raw mtim: 0x%8lx, mac2unix: 0x%8lx\n",
		otime, xtime);
	DEBUG && fflush(debug);
	fprintf(verbose, "\tModified %s", ctime((time_t *)&xtime));
	fflush(verbose);

}
